import sys
from pathlib import Path

# import rich_click as click
import click
import tomlkit
from click_rich_help import StyledGroup
from rich import box
from rich.panel import Panel
from rich.prompt import Confirm
from rich.table import Table

from ._checks import pre_run_check
from .console import console
from .extract import extract_all
from .merge import merge_all
from .read_filter import get_filter_count
from .single_cell import single_cell
from .utils import combine_outs, get_fastqs


def make_sample_check_table(
    samples, pipeline, output, quality, ratio, distance, filter, offset, queue_all
):
    processed_samples = []
    table = Table(box=box.SIMPLE, header_style="bold cyan", collapse_padding=True)
    table.add_column(
        "sample",
        justify="center",
        style="green",
        no_wrap=True,
    )
    table.add_column(f"q{quality}", justify="center")
    table.add_column("barcodes", justify="center")
    table.add_column(
        f"r{ratio}d{distance}",
        justify="center",
    )
    if "filter_count" in filter.keys():
        # filter_expr = f"min{filter['filter_count']}"
        filter_expr = "min(N)"
    else:
        # filter_expr=f"min({filter['filter_percent']}%)"
        filter_expr = "min(%)"
    table.add_column(f"{filter_expr}_off{offset}", justify="center")

    for sample in samples:
        if not queue_all:
            row = make_row(
                sample, pipeline, output, quality, ratio, distance, filter, offset
            )
            if "[bold green]" in row[0]:
                processed_samples.append(sample)
        else:
            row = [f"[bold yellow]{sample}"] + ["[yellow]Queued"] * 4

        table.add_row(*row)

    console.print(
        Panel.fit(
            table,
            title="Queue",
        )
    )

    console.print(
        f"There are {len(samples)-len(processed_samples)} samples to finish processing.\n"
    )

    return processed_samples


def check_in_dir(sample, suffix, directory):
    filepath = f"{sample}{suffix}"
    if directory / filepath in directory.glob(f"*{suffix}"):
        return "[bold green]\u2713"
    else:
        return "[yellow]Queued"


def make_row(sample, pipeline, output, quality, ratio, distance, filter, offset):

    # start the table row
    row = []
    row.append(check_in_dir(sample, f".q{quality}.fastq", pipeline))
    row.append(check_in_dir(sample, f".q{quality}.barcodes.tsv", pipeline))
    row.append(
        check_in_dir(sample, f".q{quality}.barcodes.r{ratio}d{distance}.tsv", pipeline)
    )
    if row[-1] == "[bold green]\u2713":
        if "filter_percent" in filter.keys():
            filter_count = get_filter_count(
                pipeline / f"{sample}.q{quality}.barcodes.r{ratio}d{distance}.tsv",
                filter["filter_percent"],
            )
        else:
            filter_count = filter["filter_count"]
        row.append(
            check_in_dir(
                sample,
                f".q{quality}.barcodes.r{ratio}d{distance}.min{filter_count}_off{offset}.tsv",
                output,
            )
        )

    if set(row[1:]) == {"[bold green]\u2713"}:
        row = [f"[bold green]{sample}"] + row
    else:
        row.extend(["[yellow]Queued"] * (4 - len(row)))
        row = [f"[bold yellow]{sample}"] + row

    return row


def sample_check(
    fastqs,
    pipeline,
    output,
    quality,
    ratio,
    distance,
    filter,
    offset,
):
    samples = {f.name.split(".")[0]: f for f in fastqs}
    queue_all = not pipeline.is_dir()
    processed_samples = make_sample_check_table(
        sorted(samples.keys()),
        pipeline,
        output,
        quality,
        ratio,
        distance,
        filter,
        offset,
        queue_all,
    )

    if len(processed_samples) != len(samples) and not Confirm.ask(
        "Continue with these samples?"
    ):
        sys.exit()

    return [f for sample, f in samples.items() if sample in processed_samples]


def print_params(ctx):
    params = ctx.params
    params.pop("save_config")
    params = {k: v for k, v in params.items() if v is not None}

    grid = Table.grid(expand=True)
    grid.add_column(justify="right", style="bold cyan")
    grid.add_column(justify="center")
    grid.add_column(style="yellow")

    for key, value in params.items():
        grid.add_row(key, ": ", str(value))

    console.print(
        Panel.fit(grid, title=f"{ctx.info_name.capitalize()} Parameters", padding=1)
    )


def save_params(ctx):
    cmd = ctx.info_name
    params = {k: v for k, v in ctx.params.items() if v}
    save_type = params.pop("save_config")

    try:
        config_file = Path(ctx.obj["config_file"])
    except TypeError:
        raise click.BadParameter("use `--save-config` with a specified `--config`")

    if config_file.is_file():
        console.print(f"Updating current config file at [b cyan]{config_file}")
        with config_file.open("r") as f:
            config = tomlkit.load(f)
    else:
        console.print(f"Staring a config file at [b cyan]{config_file}")
        config = tomlkit.document()

    if save_type == "explicit":
        params = {
            k: v for k, v in params.items() if ctx.get_parameter_source(k).value != 3
        }

    # sanitize the path's for writing to toml
    for k in ["input", "pipeline", "output"]:
        if k in params.keys():
            params[k] = str(params[k])

    config[cmd] = params

    null_hints = {"extract": ["filter_count", "fastp_args"], "merge": ["fastp_args"]}
    if cmd in null_hints.keys() and save_type == "full":
        for param in null_hints[cmd]:
            if param not in params.keys():
                config[cmd].add(tomlkit.comment(f"{param} ="))

    with config_file.open("w") as f:
        f.write(tomlkit.dumps(config))

    console.print("Exiting...")
    ctx.exit()


def load_params(ctx, param, filename):
    if not filename or ctx.resilient_parsing:
        return

    ctx.default_map = {}
    if Path(filename).is_file():
        with Path(filename).open("r") as f:
            params = tomlkit.load(f)
        if params:
            ctx.default_map = params.get(ctx.info_name, {})
    else:
        console.print("No config file found. ignoring..")

    ctx.obj = {"config_file": filename}


def init_check(ctx, param, check):
    if not check:
        pre_run_check()


def validate_filter_args(ctx):
    if ctx.params["filter_count"] or ctx.params["filter_count"] == 0:
        if ctx.get_parameter_source("filter_percent").value == 3:
            ctx.params["filter_percent"] = None
            del ctx.params["filter_percent"]
            return {"filter_count": ctx.params["filter_count"]}
        else:
            raise click.BadParameter(
                "`--filter-count` and `--filter-percent` are mutually exclusive"
            )
    else:
        del ctx.params["filter_count"]
        return {"filter_percent": ctx.params["filter_percent"]}


def add_options(options):
    def _add_options(func):
        for option in reversed(options):
            func = option(func)
        return func

    return _add_options


# Click option definitions
##########################


_output_options = [
    click.option(
        "-o",
        "--output",
        help="output directory",
        default="./outs",
        show_default=True,
        type=click.Path(file_okay=False, path_type=Path),
    ),
    click.option(
        "-p",
        "--pipeline",
        help="pipeline directory",
        default="./pipeline",
        show_default=True,
        type=click.Path(file_okay=False, path_type=Path),
    ),
]

_general_options = [
    click.option(
        "-v", "--verbose", help="print the output of underlying software", is_flag=True
    ),
    click.option(
        "-t",
        "--threads",
        help="number of cpu cores to use",
        default=1,
        show_default=True,
    ),
    click.option(
        "-c",
        "--config",
        type=click.Path(dir_okay=False),
        callback=load_params,
        is_eager=True,
        expose_value=False,
        help="read parameter values from config file",
        show_default=True,
    ),
    click.option(
        "--save-config",
        help="save current params to file specified by `--config`",
        type=click.Choice(["explicit", "full"], case_sensitive=False),
    ),
    click.option(
        "--skip-init-check",
        help="skip runtime dependency check",
        is_flag=True,
        hidden=True,
        expose_value=False,
        callback=init_check,
    ),
]

_trim_options = [
    click.option(
        "-e",
        "--error",
        help="error tolerance supplied to cutadapt",
        default=0.1,
        show_default=True,
    ),
    click.option(
        "-l",
        "--length",
        help="target length of extracted barcode",
        default=20,
        show_default=True,
    ),
    click.option(
        "-ua",
        "--upstream-adapter",
        help="5' sequence flanking barcode",
        default="ATCTTGTGGAAAGGACGAAACACCG",
    ),
    click.option(
        "-da",
        "--downstream-adapter",
        help="3' sequence flanking barcode",
        default="GTTTTAGAGCTAGAAATAGCAAGTT",
    ),
    click.option(
        "--unlinked-adapters", help="run cutadapt using unlinked adapters", is_flag=True
    ),
    click.option(
        "--skip-trimming",
        help="skip cutadapt trimming entirely and use reads as-is",
        is_flag=True,
    ),
]

_quality_options = [
    click.option(
        "-q",
        "--quality",
        help="minimum PHRED quality for filtering reads",
        default=30,
        show_default=True,
    ),
    click.option(
        "-up",
        "--unqualified-percent",
        help="minimum percent of bases which can be below quality threshold",
        default=20,
        show_default=True,
    ),
    click.option(
        "-fa",
        "--fastp-args",
        help="additional arguments provided as a string passed verbatim to fastp",
        type=str,
    ),
]

_cluster_options = [
    click.option(
        "-r",
        "--ratio",
        help="ratio to use for message passing clustering",
        default=3,
        show_default=True,
    ),
    click.option(
        "-d",
        "--distance",
        help="levenshtein distance for clustering",
        default=1,
        show_default=True,
        type=click.IntRange(1, 8),
    ),
]

_filter_options = [
    click.option(
        "-fc",
        "--filter-count",
        help="minium nominal number of reads",
        type=int,
    ),
    click.option(
        "-fp",
        "--filter-percent",
        help="minimum percentage of total reads",
        default=0.005,
        show_default=True,
    ),
    click.option(
        "--offset",
        help="length offset from target barcode length post-clustering",
        default=1,
        show_default=True,
    ),
]
##########################


# Click Help Groups
##########################

_general_option_group = {
    "General Options": [
        "--help",
        "--verbose",
        "--threads",
        "--config",
        "--save-config",
    ],
}

_input_output_option_group = {
    "Input/Output Options": ["--input", "--output", "--pipeline"],
}
_trim_option_group = {
    "Trim (Cutadapt) Options": [
        "--error",
        "--length",
        "--upstream-adapter",
        "--downstream-adapter",
        "--unlinked-adapters",
        "--skip-trimming",
    ],
}
_extract_option_groups = {
    **_input_output_option_group,
    **_trim_option_group,
    **{
        "Cluster (Starcode) Options": [
            "--ratio",
            "--distance",
        ],
        "Quality (Fastp) Options": [
            "--quality",
            "--unqualified-percent",
            "--fastp-args",
        ],
        "Filter Options": ["--filter-count", "--filter-percent", "--offset"],
    },
    **_general_option_group,
}


# Click Command Configs
##########################

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"], max_content_width=90)


@click.group(
    cls=StyledGroup,
    context_settings=CONTEXT_SETTINGS,
)
@click.version_option(package_name="pycashier", prog_name="pycashier")
def cli():
    """Cash in on DNA Barcode Tags
    \n\n\n
    See `[b cyan]pycashier COMMAND -h[/]` for more information.
    """
    pass


@cli.command(option_groups=_extract_option_groups)
@click.option(
    "-i",
    "--input",
    help="source directory containing fastq files",
    required=True,
    type=click.Path(exists=True, file_okay=False, path_type=Path),
)
@add_options(
    [
        *_output_options,
        *_general_options,
        *_trim_options,
        *_quality_options,
        *_cluster_options,
        *_filter_options,
    ]
)
@click.pass_context
def extract(
    ctx,
    input,
    output,
    pipeline,
    quality,
    unqualified_percent,
    fastp_args,
    skip_trimming,
    error,
    length,
    upstream_adapter,
    downstream_adapter,
    unlinked_adapters,
    ratio,
    distance,
    filter_percent,
    filter_count,
    offset,
    verbose,
    threads,
    save_config,
):
    """
    extract DNA barcodes from a directory of fastq files
    \n\n\n
    Sample names should be delimited with a ".", such as `[b cyan][yellow]<sample>[/yellow].raw.fastq[/]`,
    anything succeeding the first period will be ignored by `[b cyan]pycashier[/]`.
    \n\n\n
    If your data is paired-end with overlapping barcodes, see `[b cyan]pycashier merge[/]`.
    """

    # validate that filter count and filter percent aren't both defined
    filter = validate_filter_args(ctx)

    if save_config:
        save_params(ctx)

    console.print(("[b]\n[cyan]PYCASHIER:[/cyan] Starting Extraction\n"))
    print_params(ctx)

    fastqs = get_fastqs(input)

    processed_fastqs = sample_check(
        fastqs,
        pipeline,
        output,
        quality,
        ratio,
        distance,
        filter,
        offset,
    )

    fastqs = [f for f in fastqs if f not in processed_fastqs]

    extract_all(
        fastqs,
        output,
        pipeline,
        quality,
        unqualified_percent,
        fastp_args,
        skip_trimming,
        error,
        length,
        upstream_adapter,
        downstream_adapter,
        unlinked_adapters,
        ratio,
        distance,
        filter,
        offset,
        verbose,
        threads,
    )


@cli.command(
    option_groups={
        **_input_output_option_group,
        **{"Merge (Fastp) Options": ["--fastp-args"]},
        **_general_option_group,
    }
)
@click.option(
    "-i",
    "--input",
    help="source directory containing gzipped R1 and R2 fastq files",
    required=True,
    type=click.Path(exists=True, file_okay=False, path_type=Path),
)
@click.option(
    "-o",
    "--output",
    help="output directory",
    default="./mergedfastqs",
    show_default=True,
    type=click.Path(file_okay=False, path_type=Path),
)
@add_options([_output_options[1], *_general_options])
@click.option(
    "-fa",
    "--fastp-args",
    help="additional arguments provided as a string passed verbatim to fastp",
    type=str,
)
@click.pass_context
def merge(
    ctx,
    input,
    output,
    pipeline,
    fastp_args,
    threads,
    verbose,
    save_config,
):
    """
    merge overlapping paired-end reads using fastp
    \n\n\n
    Simple wrapper over `[b cyan]fastp[/]` to combine R1 and R2 from PE fastq files.
    \n\n\n
    [i]NOTE[/]: fastq files must be gzipped or `[b cyan]pycashier[/]` will exit.
    """

    if save_config:
        save_params(ctx)

    console.print(("[b]\n[cyan]PYCASHIER:[/cyan] Starting Merge\n"))

    print_params(ctx)

    merge_all(
        [f for f in input.iterdir()],
        input,
        pipeline,
        output,
        threads,
        verbose,
        fastp_args,
    )


@cli.command(
    option_groups={
        **_input_output_option_group,
        **{
            "Trim (Cutadapt) Options": ["--minimum-length"]
            + _trim_option_group["Trim (Cutadapt) Options"][:-2]
        },
        **_general_option_group,
    }
)
@click.option(
    "-i",
    "--input",
    help="source directory containing sam files from scRNA-seq",
    required=True,
    type=click.Path(exists=True, file_okay=False, path_type=Path),
)
@click.option(
    "-ml",
    "--minimum-length",
    help="minimum length of extracted barcode",
    default=10,
    show_default=True,
)
@add_options([*_output_options, *_trim_options[:-2], *_general_options])
@click.pass_context
def scrna(
    ctx,
    input,
    output,
    pipeline,
    minimum_length,
    length,
    error,
    upstream_adapter,
    downstream_adapter,
    threads,
    verbose,
    save_config,
):
    """
    extract expressed DNA barcodes from scRNA-seq
    \n\n\n
    Designed for interoperability with 10X scRNA-seq workflow.
    After processing samples with `[b cyan]cellranger[/]` resulting
    bam files should be converted to sam files using `[b cyan]samtools[/]`.
    \n\n\n
    [i]NOTE[/]: You can speed this up by providing a sam file with only
    the unmapped reads.
    """
    if save_config:
        save_params(ctx)

    console.print(("[b]\n[cyan]PYCASHIER:[/cyan] Starting Single Cell Extraction\n"))
    print_params(ctx)
    single_cell(
        input,
        pipeline,
        output,
        error,
        minimum_length,
        length,
        upstream_adapter,
        downstream_adapter,
        threads,
        verbose,
    )


@cli.command(
    option_groups={
        "Input/Output Options": ["--input", "--output"],
        "General Options": ["--help", "--config", "--save-config"],
    }
)
@click.option(
    "-i",
    "--input",
    help="source directory containing output files from [b cyan]pycashier extract[/]",
    default="./outs",
    show_default=True,
    type=click.Path(exists=True, file_okay=False, path_type=Path),
)
@click.option(
    "-o",
    "--output",
    help="combined tsv of all samples found in input directory",
    default="./combined.tsv",
    show_default=True,
    type=click.Path(exists=False, dir_okay=False, path_type=Path),
)
@add_options([*_general_options[2:]])
@click.pass_context
def combine(
    ctx,
    input,
    output,
    save_config,
):
    """
    combine resulting output of [b cyan]extract[/]
    """

    if save_config:
        save_params(ctx)

    combine_outs(input, output)


def main():
    cli(prog_name="pycashier")


if __name__ == "__main__":
    main()



import shlex
import shutil
import subprocess
import sys

from .console import console
from .utils import fastq_to_csv


def trim(
    sample,
    pipeline,
    fastq,
    quality,
    unqualified_percent,
    fastp_args,
    error,
    threads,
    length,
    distance,
    upstream_adapter,
    downstream_adapter,
    unlinked_adapters,
    skip_trimming,
    verbose,
):

    json_qc = pipeline / "qc" / f"{sample}.json"
    html_qc = pipeline / "qc" / f"{sample}.html"
    input_file = fastq
    filtered_fastq = pipeline / f"{sample}.q{quality}.fastq"
    filtered_barcode_fastq = pipeline / f"{sample}.q{quality}.barcode.fastq"

    if unlinked_adapters:
        adapter_string = f"-g {upstream_adapter} -a {downstream_adapter}"
    else:
        adapter_string = f"-g {upstream_adapter}...{downstream_adapter}"

    (pipeline / "qc").mkdir(exist_ok=True)

    if not filtered_fastq.is_file():

        console.print(f"[green]{sample}[/green]: quality filtering illumina reads")
        command = f"fastp \
            -i {input_file} \
            -o {filtered_fastq} \
            -q {quality} \
            -u {unqualified_percent} \
            -w {threads} \
            -h {html_qc} \
            -j {json_qc} \
            --dont_eval_duplication \
            {fastp_args or ''}"

        args = shlex.split(command)

        p = subprocess.run(
            args,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=True,
        )

        if p.returncode != 0:
            console.print("[yellow]FASTP OUTPUT:")
            console.print(f"[green]{sample}[/green]: fastp failed")
            console.print(p.stdout)
            sys.exit(1)

        elif verbose:
            console.print("[yellow]FASTP OUTPUT:")
            console.print(p.stdout)

    if skip_trimming:
        shutil.copy(filtered_fastq, filtered_barcode_fastq)

    if not filtered_barcode_fastq.is_file():

        console.print(f"[green]{sample}[/green]: extracting barcodes")

        command = f"cutadapt \
            -e {error} \
            -j {threads} \
            --minimum-length={length - distance} \
            --maximum-length={length + distance} \
            --max-n=0 \
            --trimmed-only \
            {adapter_string} \
            -n 2 \
            -o {filtered_barcode_fastq} {filtered_fastq}"

        args = shlex.split(command)

        p = subprocess.run(
            args,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=True,
        )

        if p.returncode != 0 or filtered_barcode_fastq.stat().st_size == 0:
            console.print("[yellow]CUTADAPT OUTPUT:")
            console.print(p.stdout)
            console.print(
                f"[green]{sample}[/green]: Failed to extract reads for sample"
            )
            console.print("see above for cutadapt output")
            sys.exit()

        elif verbose:
            console.print("[yellow]CUTADAPT OUTPUT:")
            console.print(p.stdout)

        console.print(f"[green]{sample}[/green]: extraction and filtering complete")

    else:
        console.print(f"[green]{sample}[/green]: skipping extraction")

    barcodes_out = pipeline / f"{sample}.q{quality}.barcodes.tsv"

    if not barcodes_out.is_file():
        console.print(f"[green]{sample}[/green]: converting fastq to tsv")
        fastq_to_csv(filtered_barcode_fastq, barcodes_out)
    else:
        console.print(f"[green]{sample}[/green]: skipping fastq to tsv conversion")



import csv

from .console import console


def get_filter_count(file_in, filter_percent):
    total_reads = 0

    with open(file_in, newline="") as csvfile:
        spamreader = csv.reader(csvfile, delimiter="\t")
        for row in spamreader:
            total_reads += float(row[1])

    filter_count = int(round(total_reads * filter_percent / 100, 0))

    return filter_count


def filter_by_percent(file_in, filter_percent, length, offset, outdir):

    filter_by_count(
        file_in, get_filter_count(file_in, filter_percent), length, offset, outdir
    )


def filter_by_count(file_in, filter_count, length, offset, output):

    name = file_in.stem
    ext = file_in.suffix
    final = output / f"{name}.min{filter_count}_off{offset}{ext}"

    with open(file_in, "r") as csv_in:
        with open(final, "w") as csv_out:
            for line in csv_in:
                linesplit = line.split("\t")
                if (
                    int(linesplit[1]) >= filter_count
                    and abs(len(linesplit[0]) - length) <= offset
                ):
                    csv_out.write(f"{linesplit[0]}\t{linesplit[1]}")

    if final.stat().st_size == 0:
        console.print(
            "[yellow]WARNING[/]: no barcodes passed final length and abundance filters"
        )


def read_filter(
    sample,
    pipeline,
    output,
    length,
    offset,
    filter,
    quality,
    ratio,
    distance,
):
    file_in = pipeline / f"{sample}.q{quality}.barcodes.r{ratio}d{distance}.tsv"

    if "filter_count" in filter.keys():
        console.print(
            f"[green]{sample}[/green]: removing "
            f"sequences with less than {filter['filter_count']} reads"
        )

        filter_by_count(file_in, filter["filter_count"], length, offset, output)

    else:
        console.print(
            f"[green]{sample}[/green]: removing"
            f" sequences with less than {filter['filter_percent']}% of total reads"
        )

        filter_by_percent(file_in, filter["filter_percent"], length, offset, output)



import sys
from pathlib import Path

from .cli import get_args, sample_check
from .cluster import cluster
from .console import console
from .extract import extract
from .merge import merge
from .read_filter import read_filter
from .single_cell import single_cell


def main():

    cli_args = get_args()

    console.rule()
    console.rule("Barcode Extraction with CASHIER")
    console.rule()

    sourcedir = Path(cli_args["main"]["sourcedir"])

    fastqs = [f for f in sourcedir.iterdir()]

    if not fastqs:
        console.print(f"Source dir: {sourcedir}, appears to be empty...")
        console.print("Exiting.")
        sys.exit(1)

    Path(cli_args["main"]["pipelinedir"]).mkdir(exist_ok=True)

    if cli_args["merge"]["merge"]:

        merge(fastqs, sourcedir, cli_args)

    Path(cli_args["main"]["outdir"]).mkdir(exist_ok=True)

    if cli_args["single_cell"]:

        single_cell(sourcedir, cli_args)

    for f in fastqs:

        ext = f.suffix

        if ext != ".fastq":
            print(
                f"ERROR! There is a non fastq file in the provided fastq directory: {f}"
            )
            print("Exiting.")
            sys.exit(1)

    processed_samples = sample_check(sourcedir, fastqs, cli_args)

    for fastq in fastqs:

        sample = fastq.name.split(".")[0]

        if sample in processed_samples:
            console.print(f"Skipping Processing for [green]{sample}[/green]")
            console.rule()
            continue

        with console.status(
            f"Processing sample: [green]{sample}[/green]", spinner="dots12"
        ):

            extract(sample, fastq, **cli_args["main"], **cli_args["extract"])

            cluster(sample, **cli_args["main"], **cli_args["cluster"])

            read_filter(
                sample,
                **cli_args["main"],
                **cli_args["filter"],
                **cli_args["cluster"],
            )

        console.print(f"[green]{sample}[/green]: processing completed")
        console.rule()

    console.print("\n[green]FINISHED!")


if __name__ == "__main__":
    main()



import shlex
import subprocess

from .console import console
from .utils import extract_csv_column


def cluster(sample, pipeline, ratio, distance, quality, threads, verbose):

    extracted_csv = pipeline / f"{sample}.q{quality}.barcodes.tsv"

    output_file = pipeline / f"{sample}.q{quality}.barcodes.r{ratio}d{distance}.tsv"

    if not output_file.is_file():
        if not extracted_csv.with_suffix(".c2.tsv").is_file():
            input_file = extract_csv_column(extracted_csv, 2)

        console.print(f"[green]{sample}[/green]: clustering barcodes")
        command = f"starcode -d {distance} -r {ratio} -t {threads} -i {input_file} -o {output_file}"

        args = shlex.split(command)

        p = subprocess.run(
            args,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=True,
        )
        if verbose:
            console.print("[yellow]STARCODE OUTPUT:")
            console.print(
                "\n".join(
                    [
                        line
                        for line in p.stdout.splitlines()
                        if not line.startswith("progress")
                    ]
                )
            )
        input_file.unlink()

        console.print(f"[green]{sample}[/green]: clustering complete")

    else:
        console.print(f"[green]{sample}[/green]: skipping clustering")



from .cluster import cluster
from .console import console
from .read_filter import read_filter
from .trim import trim


def extract_all(
    fastqs,
    output,
    pipeline,
    quality,
    unqualified_percent,
    fastp_args,
    skip_trimming,
    error,
    length,
    upstream_adapter,
    downstream_adapter,
    unlinked_adapters,
    ratio,
    distance,
    filter,
    offset,
    verbose,
    threads,
):

    for d in [pipeline, output]:
        d.mkdir(exist_ok=True)

    for fastq in sorted(fastqs):

        sample = fastq.name.split(".")[0]

        print()
        console.print(f"──────────────── {sample} ───────────────────", style="dim")

        with console.status(
            f"Processing sample: [green]{sample}[/green]", spinner="dots12"
        ):

            trim(
                sample,
                pipeline,
                fastq,
                quality,
                unqualified_percent,
                fastp_args,
                error,
                threads,
                length,
                distance,
                upstream_adapter,
                downstream_adapter,
                unlinked_adapters,
                skip_trimming,
                verbose,
            )

            cluster(sample, pipeline, ratio, distance, quality, threads, verbose)

            read_filter(
                sample,
                pipeline,
                output,
                length,
                offset,
                filter,
                quality,
                ratio,
                distance,
            )

        console.print(f"[green]{sample}[/green]: processing completed")

    console.print("\n[green]FINISHED!")



import re
import shlex
import subprocess
import sys

from rich.prompt import Confirm

from .console import console


def merge_single(sample, fastqs, input, pipeline, output, threads, verbose, fastp_args):
    (pipeline / "merge_qc").mkdir(exist_ok=True)

    # TODO: refactor for clarity and memory usage
    # also allow unzipped fastqs
    for f in fastqs:

        R1_regex = r"" + re.escape(sample) + r"\..*R1.*\.fastq.*"  # \.gz"
        m = re.search(R1_regex, f.name)
        if m:
            R1_file = m.group(0)

        R2_regex = r"" + re.escape(sample) + r"\..*R2.*\.fastq.*"  # \.gz"
        m = re.search(R2_regex, f.name)
        if m:
            R2_file = m.group(0)

    if R1_file is None or R2_file is None:
        print("oops I didn't find an R1 or R2 file")
        sys.exit(1)

    merged_barcode_fastq = output / f"{sample}.merged.raw.fastq"

    if not merged_barcode_fastq.is_file():

        path_to_r1 = input / R1_file
        path_to_r2 = input / R2_file

        console.print(f"[green]{sample}[/green]: starting fastq merge")
        command = f"fastp \
                -i {path_to_r1}  \
                -I {path_to_r2} \
                -w {threads} \
                -m -c -G -Q -L \
                -j {pipeline}/merge_qc/{sample}.json \
                -h {pipeline}/merge_qc/{sample}.html \
                --merged_out {merged_barcode_fastq} \
                {fastp_args or ''}"

        args = shlex.split(command)
        p = subprocess.run(
            args,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=True,
        )

        if p.returncode != 0 or merged_barcode_fastq.stat().st_size == 0:
            console.print("[red]FASTP FAILED!")
            console.print("[yellow]FASTP OUTPUT:")
            console.print(p.stdout)
            sys.exit()

        if verbose:
            console.print("[yellow]FASTP OUTPUT:")
            console.print(p.stdout)

    else:
        console.print(f"[green]{sample}[/green]: Found merged barcode fastq")
        console.print(f"[green]{sample}[/green]: skipping fastq merge")


def merge_all(fastqs, input, pipeline, output, threads, verbose, fastp_args):
    for d in [pipeline, output]:
        d.mkdir(exist_ok=True)

    samples = []

    for f in fastqs:

        m = re.search(r"(.+?)\..*R.*\.fastq.*", f.name)
        if m:
            samples.append(m.group(1))
        else:
            print(f"Failed to obtain sample name from {f}")
            print("Merge mode expects gzipped fastqs. Exiting.")
            sys.exit(1)

    console.print(f"[b cyan]Samples[/]: {', '.join(sorted(set(samples)))}\n")

    if not Confirm.ask("Continue with these samples?"):
        sys.exit()

    if len(samples) / len(set(samples)) != 2:
        print("There should be an R1 and R2 fastq file for each sample.")
        sys.exit(1)

    for sample in sorted(set(samples)):

        print()
        console.print(f"──────────────── {sample} ───────────────────", style="dim")

        with console.status(
            f"Processing sample: [green]{sample}[/green]", spinner="dots12"
        ):

            merge_single(
                sample, fastqs, input, pipeline, output, threads, verbose, fastp_args
            )

        console.print(f"[green]{sample}[/green]: processing completed")

    console.print("\n[green]FINISHED!")
    sys.exit()



import sys

from .console import console


def get_fastqs(src):
    fastqs = [f for f in src.iterdir()]
    if not fastqs:
        console.print(f"Source dir: {src}, appears to be empty...")
        console.print("Exiting.")
        sys.exit(1)

    for f in fastqs:

        if f.suffix not in [".fastq",".fastq.gz"] :
            print(
                f"ERROR! There is a non fastq file in the provided fastq directory: {f}"
            )
            print("Exiting.")
            sys.exit(1)

    return fastqs


def convert_to_csv(in_file, out_file):

    for i, line in enumerate(in_file):

        if not line.startswith("@") or in_file[i - 1].strip() == "+":
            continue
        seq_id = line.strip()
        sequence = in_file[i + 1].strip()
        out_file.write(f"{seq_id}\t{sequence}\n")


def fastq_to_csv(in_file, out_file):

    with open(in_file) as f_in:
        with open(out_file, "w") as f_out:
            convert_to_csv(f_in.readlines(), f_out)


def extract_csv_column(csv_file, column):

    ext = csv_file.suffix
    tmp_out = csv_file.with_suffix(f".c{column}{ext}")
    with open(csv_file, "r") as csv_in:
        with open(tmp_out, "w") as csv_out:
            for line in csv_in:
                linesplit = line.split("\t")
                csv_out.write(f"{linesplit[column-1]}")

    return tmp_out


def combine_outs(input, output):

    files = {f.name.split(".")[0]: f for f in input.iterdir()}

    console.print(f"Combing output files for {len(files)} samples.")

    with output.open("w") as csv_out:
        csv_out.write("sample\tsequence\tcount\n")
        for sample, f in files.items():
            with f.open("r") as csv_in:
                for line in csv_in:
                    csv_out.write(f"{sample}\t{line}")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pycashier']

package_data = \
{'': ['*']}

install_requires = \
['click-rich-help>=22.1.0', 'rich>=12.0.0', 'tomlkit>=0.10.1']

entry_points = \
{'console_scripts': ['pycashier = pycashier.cli:main']}

setup_kwargs = {
    'name': 'pycashier',
    'version': '0.3.3',
    'description': 'cash in on expressed barcode tags',
    'long_description': '[![Forks][forks-shield]][forks-url]\n[![Stargazers][stars-shield]][stars-url]\n[![Issues][issues-shield]][issues-url]\n[![MIT License][license-shield]][license-url]\n\n# Pycashier\n\nTool for extracting and processing DNA barcode tags from Illumina sequencing.\n\nDefault parameters are designed for use by the [Brock Lab](https://github.com/brocklab) to process data generated from\nClonMapper lineage tracing experiments, but is extensible to other similarly designed tools.\n\n\n### Bioconda Dependencies\n- [cutadapt](https://github.com/marcelm/cutadapt) (sequence extraction)\n- [starcode](https://github.com/gui11aume/starcode) (sequence clustering)\n- [fastp](https://github.com/OpenGene/fastp) (merging/quality filtering)\n- [pysam](https://github.com/pysam-developers/pysam) (sam file conversion to fastq)\n\n### Pip/conda-forge Dependencies\n- [rich](https://github.com/Textualize/rich)\n- [rich-click](https://github.com/ewels/rich-click)\n- [click-rich-help](https://github.com/daylinmorgan/click-rich-help)\n- [tomlkit](https://github.com/sdispater/tomlkit)\n\n## Installation\nIt\'s recommended to use [conda](https://docs.conda.io/en/latest/)/[mamba](https://github.com/mamba-org/mamba) to install and manage the dependencies for this package.\n\n```bash\nconda install -c conda-forge -c bioconda cutadapt fastp pysam starcode pycashier\n```\n\nYou can also use the included `environment.yml` to create your environment and install everything you need.\n\n```bash\nconda env create -f https://raw.githubusercontent.com/brocklab/pycashier/main/environment.yml\nconda activate cashierenv\n```\n\nAdditionally you may install with pip. Though it will be up to you to ensure all the\ndependencies you would install from bioconda are on your path and installed correctly.\n`Pycashier` will check for them before running.\n\n```bash\npip install pycashier\n```\n\n## Usage\n\nAs of `v0.3.0` the interface of `pycashier` has changed. Previously a positional argument was used to indicate the source directory and additional flags would set the operation.\nNow `pycashier` uses `click` and a series of commands.\n\nAs always use though use `pycashier --help` and additionally `pycashier <COMMAND> --help` for the full list of parameters.\n\nSee below for a brief explanation of each command.\n\n### Extract\n\nThe primary use case of pycashier is extracting 20bp sequences from illumina generated fastq files.\nThis can be accomplished with the below command where `./fastqs` is a directory containing all of your fastq files.\n\n```bash\npycashier extract -i ./fastqs\n```\n\n`Pycashier` will attempt to extract file names from your `.fastq` files using the first string delimited by a period.\n\nFor example:\n- `sample1.fastq`: sample1\n- `sample2.metadata_pycashier.will.ignore.fastq`: sample2\n\nAs `pycashier extract` runs, two directories will be generated `./pipeline` and `./outs`, configurable with `-p/--pipeline` and `-o/--output` respectively.\n\nYour `pipeline` directory will contain all files and data generated while performing barcode extraction and clustering.\nWhile `outs` will contain a single `.tsv` for each sample with the final barcode counts.\n\nExpected output of `pycashier extract`:\n\n```bash\nfastqs\n└── sample.raw.fastq\npipeline\n├── qc\n│\xa0\xa0 ├── sample.html\n│\xa0\xa0 └── sample.json\n├── sample.q30.barcode.fastq\n├── sample.q30.barcodes.r3d1.tsv\n├── sample.q30.barcodes.tsv\n└── sample.q30.fastq\nouts\n└── sample.q30.barcodes.r3d1.min176_off1.tsv\n```\n\n*NOTE*: If you wish to provide `pycashier` with fastq files containing only your barcode you can supply the `--skip-trimming` flag.\n\n### Merge\n\nIn some cases your data may be from paired-end sequencing. If you have two fastq files per sample\nthat overlap on the barcode region they can be combined with `pycashier merge`.\nthat overlap on the barcode region they can be combined with `pycashier merge`.\n\n\n```bash\npycashier merge -i ./fastqgz\n```\n\nBy default your output will be in `mergedfastqs`. Which you can then pass back to `pycashier` with `pycashier extract -i mergedfastqs`.\n\nFor single read, files are `<sample>.fastq` now they should both contain R1 and R2 and additionally may be gzipped.\n\nFor example:\n- `sample.raw.R1.fastq.gz`,`sample.raw.R2.fastq.gz`: sample\n- `sample.R1.fastq`,`sample.R2.fastq`: sample\n- `sample.fastq`: fail, not R1 and R2\n\n\n### Scrna\n\nIf your DNA barcodes are expressed and detectable in 10X 3\'-based transcriptomic sequencing,\nthen you can extract these tags with `pycashier` and their associated umi/cell barcodes from the `cellranger` output.\n\nFor `pycashier scrna` we extract our reads from sam files.\nThis file can be generated using the output of `cellranger count`.\nFor each sample you would run:\n```\nsamtools view -f 4 $CELLRANGER_COUNT_OUTPUT/sample1/outs/possorted_genome_bam.bam > sams/sample1.unmapped.sam\n```\nThis will generate a sam file containing only the unmapped reads.\n\nThen similar to normal barcode extraction you can pass a directory of these unmapped sam files to pycashier and extract barcodes. You can also still specify extraction parameters that will be passed to cutadapt as usual.\n\n*Note*: The default parameters passed to cutadapt are unlinked adapters and minimum barcode length of 10 bp.\n\n```\npycashier scrna -i sams\n```\n\nWhen finished the `outs` directory will have a `.tsv` containing the following columns: Illumina Read Info, UMI Barcode, Cell Barcode, gRNA Barcode\n\n### Combine\n\nThis command can be used if you wish to generate a combined tsv from all files including headers and sample information.\nBy default it uses `./outs` for input and `./combined.tsv` for output.\n\n## Config File\n\nAs of `v0.3.1` you may generate and supply `pycashier` with a toml config file using `-c/--config`.\nThe expected structure is each command followed by key value pairs of flags with hypens replaced by underscores:\n\n```toml\n[extract]\ninput = "fastqs"\nthreads = 10\nunqualified_percent = 20\n\n[merge]\ninput = "rawfastqgzs"\noutput = "mergedfastqs"\nfastp_args = "-t 1"\n```\n\nThe order of precedence for arguments is command line > config file > defaults.\n\nFor example if you were to use the above `config.toml` with `pycashier extract -c config.toml -t 15`.\nThe value used for threads would be 15.\nYou can confirm the parameter values as they will be printed prior to any execution.\n\nFor convenience, you can update/create your config file with `pycasher COMMAND --save-config [explicit|full] -c config.toml`.\n\n"Explicit" will only save parameters already included in the config file or specified at runtime.\n"Full" will include all parameters again maintaining preset values in config or specified at runtime.\n\n## Non-Configurable Defaults\n\nSee below for the non-configurable flags provided to external tools in each command. Refer to their documentation regarding the purpose of these flags.\n\n### Extract\n\n- `fastp`: `--dont_eval_duplication`\n- `cutadapt`: `--max-n=0 -n 2`\n\n### Merge\n\n- `fastp`: `-m -c -G -Q -L`\n\n### Scrna\n\n- `cutadapt`: `--max-n=0 -n 2`\n\n## Usage notes\nPycashier will **NOT** overwrite intermediary files. If there is an issue in the process, please delete either the pipeline directory or the requisite intermediary files for the sample you wish to reprocess. This will allow the user to place new fastqs within the source directory or a project folder without reprocessing all samples each time.\n- If there are reads from multiple lanes they should first be concatenated with `cat sample*R1*.fastq.gz > sample.R1.fastq.gz`\n- Naming conventions:\n    - Sample names are extracted from files using the first string delimited with a period. Please take this into account when naming sam or fastq files.\n    - Each processing step will append information to the input file name to indicate changes, again delimited with periods.\n\n\n## Acknowledgments\n\n[Cashier](https://github.com/russelldurrett/cashier) is a tool developed by Russell Durrett for the analysis and extraction of expressed barcode tags.\nThis version like it\'s predecessor wraps around several command line bioinformatic tools to pull out expressed barcode tags.\n\n\n\n[forks-shield]: https://img.shields.io/github/forks/brocklab/pycashier.svg?style=flat\n[forks-url]: https://github.com/brocklab/pycashier/network/members\n[stars-shield]: https://img.shields.io/github/stars/brocklab/pycashier.svg?style=flat\n[stars-url]: https://github.com/brocklab/pycashier/stargazers\n[issues-shield]: https://img.shields.io/github/issues/brocklab/pycashier.svg?style=flat\n[issues-url]: https://github.com/brocklab/pycashier/issues\n[license-shield]: https://img.shields.io/github/license/brocklab/pycashier.svg?style=flat\n[license-url]: https://github.com/brocklab/pycashier/blob/main/LICENSE\n',
    'author': 'Daylin Morgan',
    'author_email': 'daylinmorgan@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/brocklab/pycashier/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<3.10',
}


setup(**setup_kwargs)
